abstract sig Node{domains: set Domain, id: set NodeID, demux: set Demux}
sig Client extends Node{want: set Keyword}{#domains=1 && #id=1 && #demux>=1}
sig GW extends Node{
	state: (NodeID set -> set Demux) set -> set Demux
	}
	{#domains=2 && 
	(no disj d1,d2: Domain| (d1 in IPdomain && d2 in IPdomain) || 
	(d1 in NDNdomain && d2 in NDNdomain) || (d1 in MFdomain && d2 in MFdomain))
	&& #id=2 && #demux>=2}
sig ORS extends Node{map: Keyword one -> one ContentID}
	{#domains=1 && #id=1 && #demux>=1}

abstract sig NodeID{}
sig IPaddress extends NodeID{}
sig NDNclientID extends NodeID{}
sig NodeGUID extends NodeID{}

abstract sig Domain{}
sig IPdomain extends Domain{}
sig NDNdomain extends Domain{}
sig MFdomain extends Domain{}

abstract sig Demux{}
sig Port extends Demux{}
sig NDNreqID extends Demux{}
sig MFreqID extends Demux{}

sig Keyword{}
sig ContentID{}

abstract sig Route{initiator, acceptor: Node, demux: Demux, domain: one Domain, keyword: Keyword, clientdomain: Domain, orsdomain: Domain}
	{initiator!=acceptor && domain in initiator.domains && domain in acceptor.domains && demux in initiator.demux}
sig IProute extends Route{srcIP, dstIP: IPaddress, dstPort: Port}
	{domain in IPdomain && srcIP in initiator.id && dstIP in acceptor.id && dstPort in acceptor.demux
	&& demux in Port}
sig NDNroute extends Route{clientID: NDNclientID} 
	{domain in NDNdomain && clientID in initiator.id && demux in NDNreqID}
sig MFroute extends Route{srcGUID, dstGUID: NodeGUID}
	{domain in MFdomain && srcGUID in initiator.id && dstGUID in acceptor.id
	&& demux in MFreqID}

abstract sig RouteR{initiator, acceptor: Node, demux: Demux, domain: one Domain, contentID:ContentID}
	{initiator!=acceptor && domain in initiator.domains && domain in acceptor.domains && demux in acceptor.demux}
sig IProuteR extends RouteR{srcIP, dstIP: IPaddress, srcPort: Port}
	{domain in IPdomain && srcIP in initiator.id && dstIP in acceptor.id && srcPort in initiator.demux
	&& demux in Port}
sig NDNrouteR extends RouteR{clientID: NDNclientID} 
	{domain in NDNdomain && clientID in acceptor.id && demux in NDNreqID}
sig MFrouteR extends RouteR{srcGUID, dstGUID: NodeGUID}
	{domain in MFdomain && srcGUID in initiator.id && dstGUID in acceptor.id
	&& demux in MFreqID}

one sig Connections{connected: Route -> Route, connectedR: RouteR -> RouteR}

fact NodeIDdomain{
	all n:Node, d:Domain |
		d in n.domains iff
			some nid:NodeID| nid in n.id &&
					(nid in IPaddress iff d in IPdomain) &&
					(nid in NDNclientID iff d in NDNdomain) &&
					(nid in NodeGUID iff d in MFdomain)
}

fact demux{
	all n:Node, d:Domain |
		d in n.domains iff
			some dx:Demux | dx in n.demux &&
				(dx in Port iff d in IPdomain) &&
				(dx in NDNreqID iff d in NDNdomain) &&
				(dx in MFreqID iff d in MFdomain)
}


fact connected{
	all r1,r2:Route , c:Connections |
		(r1->r2) in c.connected iff
			r1.acceptor = r2.initiator && r1.keyword=r2.keyword && r1.clientdomain=r2.clientdomain
			&& r1.orsdomain=r2.orsdomain
	all r1,r2:RouteR , c:Connections |
		(r1->r2) in c.connectedR iff
			r1.acceptor = r2.initiator && r1.contentID=r2.contentID && 
			(some gw:GW| r1.acceptor=gw && r2.initiator=gw =>
				some n:NodeID, d1,d2:Demux| (n->d1->d2) in gw.state && n in r2.acceptor.id 
					&& d1 in r2.acceptor.demux && d2 in r1.acceptor.demux && d1=r2.demux && d2=r1.demux
					)
}

fact pathexists{
	all co:Connections, disj n1,n2:Node, k:Keyword, cd:Domain, od: Domain| 
		(some c:Client, o:ORS| cd in c.domains && od in o.domains =>
		(some r1,r2:Route | 
		(r1->r2) in ^(co.connected) && r1.initiator=n1 && r2.acceptor=n2 && 
		r1.keyword=k && r2.keyword=k && r1.clientdomain=cd && r2.clientdomain=cd &&
		r1.orsdomain=od && r2.orsdomain=od))

	all co:Connections, disj n1,n2:Node, conid:ContentID, cd:Domain, od:Domain| 
		(some c:Client, o:ORS| cd in c.domains && od in o.domains =>
		(some r1,r2:RouteR | 
		(r1->r2) in ^(co.connectedR) && r1.initiator=n1 && r2.acceptor=n2 && 
		r1.contentID=conid && r2.contentID=conid))

}

fact NDNReversePath{
	all rr:NDNrouteR| 
		some r:NDNroute, o:ORS|(r.keyword -> rr.contentID) in o.map =>
				rr.initiator=r.acceptor && rr.acceptor=r.initiator && rr.domain=r.domain
				&& rr.contentID=o.map[r.keyword] && r.demux=rr.demux && r.clientID=rr.clientID
}

fact{
	//#ORS=1 #Client=1
	all o:ORS, c:Client| no d:Domain| d in o.domains && d in c.domains
	all d:Domain| some gw:GW| d in gw.domains
	//#GW=2
	all nid:NodeID | no disj n1,n2:Node | nid in n1.id && nid in n2.id
	//all dx:Demux | no disj n1,n2:Node | dx in n1.demux && dx in n2.demux
	//#Keyword=2
	all disj r1,r2:Route| r1.initiator=r2.initiator => r1.demux!=r2.demux
	all disj r1,r2:RouteR| r1.initiator=r2.initiator => r1.demux!=r2.demux
	//#Domain=3
	#Client.want>=1
	//#Route>4
	//#RouteR>4
	//#NDNroute>=1 && #MFroute>=1 && #IProute>=1
	//#NDNrouteR>=1 && #MFrouteR>=1 && #IProuteR>=1
	//#GW.state>=1
}

//pred empty{}
//run empty for 6

pred reach[c:Client, k:Keyword,o:ORS,gw:GW]{
	all co: Connections| k in c.want =>
		(some r:Route, cid:ContentID| r.initiator=c && r.acceptor=gw && r.keyword=k
		&& r.demux in c.demux
		=> some r1,r2:Route | (r1->r2) in ^(co.connected) && r1.initiator=gw && r2.acceptor=o
		&& r1.demux in gw.demux
		&& r1.keyword=k && r2.keyword=k && (k->cid) in o.map
		&& r1.clientdomain in c.domains && r2.clientdomain in c.domains
		&& r1.orsdomain in o.domains && r2.orsdomain in o.domains
		)
}
run reach for 1 Client, 2 GW, 1 ORS, 6 NodeID, 3 Domain, 6 Port, 3 NDNreqID, 3 MFreqID, 1 Keyword, 1 ContentID, 12 Route, 1 Connections, 12 RouteR


assert reach{
	all c:Client, k:Keyword| some o:ORS, gw:GW| reach[c,k,o,gw]
}
check reach for 1 Client, 2 GW, 1 ORS, 6 NodeID, 3 Domain, 6 Port, 3 NDNreqID, 3 MFreqID, 1 Keyword, 1 ContentID, 12 Route, 1 Connections, 12 RouteR


pred return[c:Client, k:Keyword, o:ORS, gw:GW]{
	all co:Connections| some gw1:GW | reach[c,k,o,gw1] =>
		(some r,r1,r2:RouteR| (r1->r2) in ^(co.connectedR) && r1.initiator=o && r2.acceptor=gw && r2.demux in gw.demux
			&& r1.contentID=o.map[k] && r2.contentID=o.map[k] && r.initiator=gw && r.acceptor=c 
			&& r.demux in c.demux && r.contentID=o.map[k]
		)

}

run return for 1 Client, 2 GW, 1 ORS, 6 NodeID, 3 Domain, 6 Port, 3 NDNreqID, 3 MFreqID, 1 Keyword, 1 ContentID, 12 Route, 1 Connections, 12 RouteR

assert return{
	all c:Client, k:Keyword, o:ORS| some gw:GW| return[c,k,o,gw]

}

check return for 1 Client, 2 GW, 1 ORS, 6 NodeID, 3 Domain, 6 Port, 3 NDNreqID, 3 MFreqID, 1 Keyword, 1 ContentID, 12 Route, 1 Connections, 12 RouteR

pred noconflict1[c:Client, disj k1,k2:Keyword]{//same client
//{	all c:Client, k1,k2:Keyword | //noconflict1[c,k1,k2]
	some o1,o2:ORS, gw1,gw2:GW | 
		return[c,k1,o1,gw1] && return[c,k2,o2,gw2] =>
			some n1,n2: NodeID, d1,d2,d3,d4:Demux| (n1->d1->d2) in gw1.state && (n2->d3->d4) in gw2.state &&
				n1 in c.id && d1 in c.demux && d2 in gw1.demux && n2 in c.id && d3 in c.demux && d4 in gw2.demux &&
				!(n1=n2 && d1=d3 && d2=d4) &&
				(some disj r1,r2:RouteR| 
					r1.initiator=gw1 && r1.acceptor=c && r1.contentID=o1.map[k1] && r1.demux=d1 &&
					r2.initiator=gw2 && r2.acceptor=c && r2.contentID=o2.map[k2] && r2.demux=d3
				)
}

run noconflict1 for 1 Client, 1 GW, 1 ORS, 4 NodeID, 2 Domain, 4 Port, 4 NDNreqID, 4 MFreqID, 2 Keyword, 2 ContentID, 8 Route, 1 Connections, 8 RouteR

assert noconflict1{
	all c:Client, disj k1,k2:Keyword | noconflict1[c,k1,k2]
}

check noconflict1 for 1 Client, 1 GW, 1 ORS, 4 NodeID, 2 Domain, 4 Port, 4 NDNreqID, 4 MFreqID, 2 Keyword, 2 ContentID, 8 Route, 1 Connections, 8 RouteR

pred noconflict2[c1,c2:Client, k:Keyword]{//same keyword, different clients
//{	all c:Client, k1,k2:Keyword | //noconflict1[c,k1,k2]
	some o1,o2:ORS, gw1,gw2:GW | 
		return[c1,k,o1,gw1] && return[c2,k,o2,gw2] =>
			some n1,n2: NodeID, d1,d2,d3,d4:Demux| (n1->d1->d2) in gw1.state && (n2->d3->d4) in gw2.state &&
				n1 in c1.id && d1 in c1.demux && d2 in gw1.demux && n2 in c2.id && d3 in c2.demux && d4 in gw2.demux &&
				!(n1=n2 && d1=d3 && d2=d4) &&
				(some disj r1,r2:RouteR| 
					r1.initiator=gw1 && r1.acceptor=c1 && r1.contentID=o1.map[k] && r1.demux=d1 &&
					r2.initiator=gw2 && r2.acceptor=c2 && r2.contentID=o2.map[k] && r2.demux=d3
				)
}

run noconflict2 for 2 Client, 1 GW, 1 ORS, 5 NodeID, 2 Domain, 4 Port, 4 NDNreqID, 4 MFreqID, 1 Keyword, 1 ContentID, 8 Route, 1 Connections, 8 RouteR

assert noconflict2{
	all c1,c2:Client, k:Keyword| noconflict2[c1,c2,k]
}

check noconflict2 for 2 Client, 1 GW, 1 ORS, 5 NodeID, 2 Domain, 4 Port, 4 NDNreqID, 4 MFreqID, 1 Keyword, 1 ContentID, 8 Route, 1 Connections, 8 RouteR

pred NDNgw[c:Client, k:Keyword, o:ORS, gw1,gw2: GW]{
	reach[c,k,o,gw1] && return[c,k,o,gw2] && (some n:NDNdomain | n in c.domains) 
		=>
			gw1=gw2
}

run NDNgw for 1 Client, 2 GW, 1 ORS, 6 NodeID, 3 Domain, 6 Port, 3 NDNreqID, 3 MFreqID, 1 Keyword, 1 ContentID, 12 Route, 1 Connections, 12 RouteR

assert NDNgw{
	all c:Client, k:Keyword, o:ORS | some gw1,gw2:GW | NDNgw[c,k,o,gw1,gw2]
}

check NDNgw for 1 Client, 2 GW, 1 ORS, 6 NodeID, 3 Domain, 6 Port, 3 NDNreqID, 3 MFreqID, 1 Keyword, 1 ContentID, 12 Route, 1 Connections, 12 RouteR
