
one sig Domain{
	t_event,d_rereg, d_regpop, d_recover, d_retransmit, d_transmit: Int,
	binding: Binding
}

abstract sig Binding{}
sig LateBinding extends Binding{}
sig EarlyBinding extends Binding{}

abstract sig Mobility{}
sig Stationary extends Mobility{}
sig Mobile extends Mobility{}

sig Name{}
sig Address{}
//sig Time{value:Int}
one sig Bind{
//	map: Name -> Address -> Time
}
//{
//	all n:Name, a:Address | one t:Time | (n->a->t) in map
//}
sig Route{
	initiator: Sender, acceptor: Receiver,
	destination: Address,	name: Name,
	t_birth, t_death: Int
}
{
	destination in acceptor.address
	name in acceptor.name
	t_death>t_birth
	//(name -> destination -> t_birth) in Bind.map
}

abstract sig Agent{}
sig Sender extends Agent{
	t_begin, t_end: Int
}{t_begin<t_end && t_end.sub[t_begin] >= Domain.d_transmit}
sig Receiver extends Agent{address: some Address, name:Name, mobility:Mobility}

fact{
	#Route=2
	all disj r1,r2:Route| r1.name=r2.name && r1.acceptor=r2.acceptor 
		&& r1.destination not in r2.destination
	no addr:Address, disj a1,a2:Receiver | addr in a1.address & a2.address
	Domain.t_event = 10
	1 in Domain.d_rereg
	1 in Domain.d_regpop
	1 in Domain.d_recover
	1 in Domain.d_retransmit
	Domain.d_transmit>= 0 && Domain.d_transmit<=10 // range: 1 till 10, 15, or 20
	//Domain.binding in LateBinding
	Sender.t_begin = 5
	Sender.t_end	= 25
	one r:Route | r.t_birth=0 && r.t_death in Domain.t_event
	one r:Route| some d:Domain| r.t_birth = d.t_event.add[d.d_rereg].add[d.d_regpop] && r.t_death=30
	//one r:Route| r.t_birth=-1 && r.t_death=30

}


pred stationary[snd:Sender, rcv:Receiver, n:Name]{
	some r:Route| r.initiator=snd && r.acceptor=rcv && r.name=n && r.destination in rcv.address
		//&& (r.name->r.destination) in Bind.map
		&& r.destination in rcv.address 
		&& snd.t_begin.add[Domain.d_transmit]<r.t_death 
		&& snd.t_begin.add[Domain.d_transmit]<snd.t_end

}
run stationary for 1 Domain,  1 Name, 2 Address, 2 Route, 1 Receiver, 1 Sender, 7 Int, 1 Binding, 1 Mobility

pred mobileLateBinding[snd:Sender, rcv:Receiver, n:Name]{
	some disj r1,r2: Route| r1.initiator=snd && r2.initiator=snd && r1.acceptor=rcv && r2.acceptor=rcv
 		&& r1.name=n && r2.name=n 
		//&& (r1.name->r1.destination) in Bind.map &&  (r2.name->r2.destination) in Bind.map
		&& r1.destination in rcv.address && r2.destination in rcv.address
		&& r1.t_death.sub[snd.t_begin].add[r2.t_death].sub[r2.t_birth].sub[Domain.d_recover] >= Domain.d_transmit
		&& r1.t_death.sub[snd.t_begin].add[snd.t_end].sub[r2.t_birth].sub[Domain.d_recover] >= Domain.d_transmit
		&& r1.t_death.sub[snd.t_begin].add[r2.t_death].sub[r2.t_birth].sub[Domain.d_recover].add[Domain.d_rereg].add[Domain.d_regpop]<snd.t_end 		
}
run mobileLateBinding for 1 Domain,  1 Name, 2 Address, 2 Route, 1 Receiver, 1 Sender, 7 Int, 1 Binding, 1 Mobility

pred mobileEarlyBinding[snd:Sender, rcv:Receiver, n:Name]{
	some disj r1,r2: Route| r1.initiator=snd && r2.initiator=snd && r1.acceptor=rcv && r2.acceptor=rcv
 		&& r1.name=n && r2.name=n 
		//&& (r1.name->r1.destination) in Bind.map &&  (r2.name->r2.destination) in Bind.map
		&& r1.destination in rcv.address && r2.destination in rcv.address
		&& r2.t_death.sub[r2.t_birth].sub[Domain.d_retransmit] >= Domain.d_transmit
		&& snd.t_end.sub[r2.t_birth].sub[Domain.d_retransmit] >= Domain.d_transmit
		&& r2.t_death.sub[r2.t_birth].sub[Domain.d_retransmit].add[Domain.d_rereg].add[Domain.d_regpop]<snd.t_end
		&& r2.t_death.sub[snd.t_begin]>0 		
}
run mobileEarlyBinding for 1 Domain,  1 Name, 2 Address, 2 Route, 1 Receiver, 1 Sender, 7 Int, 1 Binding, 1 Mobility

//pred reach[snd:Sender, rcv:Receiver, n:Name]{
//	(stationary[snd,rcv,n] and rcv.mobility in Stationary)
//	or (mobileLateBinding[snd,rcv,n] and rcv.mobility in Mobility and Domain.binding in LateBinding) 
//	or (mobileEarlyBinding[snd,rcv,n] and rcv.mobility in Mobility and Domain.binding in EarlyBinding) 
//}

//run reach for 1 Domain,  1 Name, 2 Address, 2 Route, 1 Receiver, 1 Sender, 7 Int, 1 Binding, 1 Mobility

assert stationary{
all snd:Sender, rcv:Receiver, n:Name| stationary[snd, rcv, n]}

check stationary for 1 Domain,  1 Name, 2 Address, 2 Route, 1 Receiver, 1 Sender, 7 Int, 1 Binding, 1 Mobility

assert mobileLateBinding{all snd:Sender, rcv:Receiver, n:Name| mobileLateBinding[snd, rcv, n]}

check mobileLateBinding for 1 Domain,  1 Name, 2 Address, 2 Route, 1 Receiver, 1 Sender, 7 Int, 1 Binding, 1 Mobility

assert mobileEarlyBinding{all snd:Sender, rcv:Receiver, n:Name| mobileEarlyBinding[snd, rcv, n]}

check mobileEarlyBinding for 1 Domain,  1 Name, 2 Address, 2 Route, 1 Receiver, 1 Sender, 7 Int, 1 Binding, 1 Mobility

//assert reach{all snd:Sender, rcv:Receiver, n:Name| reach[snd, rcv, n]}

//check reach for 1 Domain,  1 Name, 2 Address, 2 Route, 1 Receiver, 1 Sender, 7 Int, 1 Binding, 1 Mobility
