abstract sig Node{domains: set Domain}
sig Client extends Node{want: set Prefix}{#domains=1}
sig GW extends Node{state: Prefix set -> set GroupID}
	{#domains=2 && 
	(no disj d1,d2: Domain| (d1 in domains && d2 in domains) && (d1 in IPdomain && d2 in IPdomain) || 
	(d1 in NDNdomain && d2 in NDNdomain) || (d1 in MFdomain && d2 in MFdomain))}
sig Publisher extends Node{map: ContentID one -> one Content}{#domains=1}

abstract sig Domain{}
sig IPdomain extends Domain{}
sig NDNdomain extends Domain{}
sig MFdomain extends Domain{}

sig Route{initiator, acceptor: Node, domain: Domain}
	{initiator!=acceptor && domain in initiator.domains && domain in acceptor.domains}


sig Content{}
abstract sig ContentID{prefix: Prefix}
sig URL extends ContentID{}
sig Name extends ContentID{}
sig ContentGUID extends ContentID{}

one sig Connection{connected: Route->Route, connectedR: Route->Route, 
	chain:Group->Group, chainR:Group->Group}

sig Prefix{parent: lone Prefix, domains: some Domain}{}

one sig PTree {map: Prefix set -> set Prefix}

sig Group{domain: Domain, gid: lone GroupID, prefix: Prefix}
	{(domain in IPdomain => #gid=1 && gid in MIPaddress) &&
	  (domain in MFdomain => #gid=1 && gid in MGUID) &&
	   (domain in NDNdomain => #gid=0)}

abstract sig GroupID{}
sig MIPaddress extends GroupID{}
sig MGUID extends GroupID{}

fact path{
	all r1,r2:Route, co:Connection | 
		(r1->r2) in co.connected <=> (r1.acceptor=r2.initiator)
	all n1,n2:Node, co:Connection| some r1,r2:Route| r1.initiator=n1 && r2.acceptor=n2
		&& (r1->r2) in ^(co.connected)

			
	all n1,n2:Node, co:Connection| some r1,r2:Route| r1.initiator=n1 && r2.acceptor=n2
		&& (r1->r2) in ^(co.connected)

}

fact PrefixRules{
	all p:Prefix| p.parent!= p
	all p1,p2:Prefix, pt:PTree| p1 in p2.parent <=> (p1->p2) in pt.map
	one p:Prefix| #p.parent=0
	all pt:PTree| no p:Prefix| (p->p) in ^(pt.map)
	all disj p1,p2: Prefix, pt:PTree | #p1.parent=0 => (p1->p2) in ^(pt.map)		
}

fact GroupRules{
	all disj g1,g2:Group, co:Connection | 
		(g1->g2) in co.chain <=> 
			(g1.prefix=g2.prefix && 
			(some gw:GW| g1.domain in gw.domains && g2.domain in gw.domains))

	all disj d1,d2:Domain, co:Connection, p:Prefix|
		some disj g1,g2:Group |
			g1.domain=d1 && g2.domain=d2 && (g1->g2) in ^(co.chain) 
			&& g1.prefix=p && g2.prefix=p
//R
	all disj g1,g2:Group, co:Connection | 
		(g1->g2) in co.chainR <=> 
			(g1.prefix=g2.prefix && 
			some gw:GW| g1.domain in gw.domains && g2.domain in gw.domains
			&& (g1.prefix->g2.gid) in gw.state)	
	all disj d1,d2:Domain, co:Connection, p:Prefix|
		some disj g1,g2:Group |
			g1.domain=d1 && g2.domain=d2 && (g1->g2) in ^(co.chainR) 
			&& g1.prefix=p && g2.prefix=p		
}

fact{
	#Client=1 #GW=2 #Publisher=1 #Domain>=3
	#Client.want>=1
	all pub:Publisher, c:Client| no d:Domain| d in pub.domains && d in c.domains
	all d:Domain| some gw:GW| d in gw.domains
	//all ndn:NDNdomain, c:Client, pub:Publisher | c.domains in ndn or pub.domains in ndn
	//#NDNdomain <=2
	#Route=6
	#Prefix=3 #Name=3
	all p:Prefix, d:Domain | one g:Group| g.domain=d && g.prefix=p
	#GroupID = #Group
	all disj g1,g2:Group | no grid:GroupID| grid in g1.gid && grid in g2.gid  
}

pred empty{}
run empty for 1 Client, 1 Publisher, 2 GW, 3 Domain, 12 Route, 3 Prefix, 3 Content, 3 ContentID, 
	1 Connection, 1 PTree, 9 Group, 9 GroupID


pred sub[c:Client, p:Prefix]{
	all co:Connection |
		p in c.want => (
		(some pub:Publisher, cid:ContentID, con:Content| (cid->con) in pub.map && cid.prefix=p
		=> (some r1,r2:Route| r1.initiator=c && r2.acceptor=pub && (r1->r2) in ^(co.connected) &&
			some g1,g2:Group| g1.domain=c.domain && g2.domain=pub.domain
				&& g1.prefix=p && g2.prefix=p && (g1->g2) in ^(co.chain))))
			
}

run sub for 1 Client, 1 Publisher, 2 GW, 3 Domain, 12 Route, 3 Prefix, 3 Content, 3 ContentID, 
	1 Connection, 1 PTree, 9 Group, 9 GroupID

assert sub{
	all c:Client, p:Prefix| sub[c,p]
}

check sub for 1 Client, 1 Publisher, 2 GW, 3 Domain, 12 Route, 3 Prefix, 3 Content, 3 ContentID, 
	1 Connection, 1 PTree, 9 Group, 9 GroupID

pred rcv[pub:Publisher, con:Content, cid:ContentID]{
	all co:Connection|
		(cid->con) in pub.map => (
		(some c:Client, p:Prefix| p in c.want && cid.prefix=p
		=> (some r1,r2:Route| r1.initiator=pub && r2.acceptor=c && (r1->r2) in ^(co.connected) &&
			some g1,g2:Group| g1.domain=pub.domain && g2.domain=c.domain
				&& g1.prefix=p && g2.prefix=p && (g1->g2) in ^(co.chainR))))	
}

run rcv for 1 Client, 1 Publisher, 2 GW, 3 Domain, 12 Route, 3 Prefix, 3 Content, 3 ContentID, 
	1 Connection, 1 PTree, 9 Group, 9 GroupID

assert rcv{
	all pub:Publisher, con:Content, cid:ContentID | rcv[pub,con,cid]
}

check rcv for 1 Client, 1 Publisher, 2 GW, 3 Domain, 12 Route, 3 Prefix, 3 Content, 3 ContentID, 
	1 Connection, 1 PTree, 9 Group, 9 GroupID

pred rcvall[pub:Publisher, con:Content, cid:ContentID]{
	all co:Connection, pt:PTree|
		(cid->con) in pub.map and pub.domains in NDNdomain => (
		(some c:Client, p:Prefix| (p in c.want or (all p1:Prefix| (p1->p) in ^(pt.map) && p1 in c.want)) && cid.prefix=p
		=> (some r1,r2:Route| r1.initiator=pub && r2.acceptor=c && (r1->r2) in ^(co.connected) &&
			some g1,g2:Group| g1.domain=pub.domain && g2.domain=c.domain
				&& g1.prefix=p && g2.prefix=p && (g1->g2) in ^(co.chainR))))	
			
}

run rcvall for 1 Client, 1 Publisher, 2 GW, 3 Domain, 12 Route, 3 Prefix, 3 Content, 3 ContentID, 
	1 Connection, 1 PTree, 9 Group, 9 GroupID


assert  rcvall{
	all pub:Publisher, con:Content, cid:ContentID| rcvall[pub, con, cid]	
			
}

check rcvall for 1 Client, 1 Publisher, 2 GW, 3 Domain, 12 Route, 3 Prefix, 3 Content, 3 ContentID, 
	1 Connection, 1 PTree, 9 Group, 9 GroupID
