abstract sig Node{domains: set Domain, id: set NodeID}
sig Client extends Node{want: set ContentID}{#domains=1 && #id=1}
sig GW extends Node{
	state1: NodeID set -> set ContentID,
	state2: set ContentID,
	state3: ContentID set -> set ContentID	
	}
	{#domains=2 && 
	(no disj d1,d2: Domain| (d1 in IPdomain && d2 in IPdomain) || 
	(d1 in NDNdomain && d2 in NDNdomain) || (d1 in MFdomain && d2 in MFdomain))
	&& #id=2}
sig Repos extends Node{map: ContentID one -> one Content}
	{#domains=1 && #id=1 &&
	(some ipd:IPdomain| ipd in domains => (all cid:ContentID, c:Content| (cid->c) in map => cid in URL)) &&
	(some ndnd:IPdomain| ndnd in domains => (all cid:ContentID, c:Content| (cid->c) in map => cid in Name)) &&
	(some mfd:IPdomain| mfd in domains => (all cid:ContentID, c:Content| (cid->c) in map => cid in ContentGUID)) 
	}

abstract sig NodeID{}
sig IPaddress extends NodeID{}
sig NDNclientID extends NodeID{}
sig NodeGUID extends NodeID{}

abstract sig Domain{}
sig IPdomain extends Domain{}
sig NDNdomain extends Domain{}
sig MFdomain extends Domain{}

sig Content{}
abstract sig ContentID{}
sig URL extends ContentID{}
sig Name extends ContentID{}
sig ContentGUID extends ContentID{}

abstract sig Route{initiator, acceptor: Node, domain: one Domain, contentID: ContentID, reposdomain: Domain}
	{initiator!=acceptor && domain in initiator.domains && domain in acceptor.domains}
sig IProute extends Route{srcIP, dstIP: IPaddress}
	{domain in IPdomain && srcIP in initiator.id && dstIP in acceptor.id}
sig NDNroute extends Route{} 
	{domain in NDNdomain}
sig MFroute extends Route{srcGUID: NodeGUID, dstGUID: lone NodeGUID}
	{domain in MFdomain && srcGUID in initiator.id && dstGUID in acceptor.id &&
	(some n:Node| n.id=dstGUID && n not in Repos => #dstGUID=1)}

abstract sig RouteR{initiator, acceptor: Node, domain: one Domain, content:Content}
	{initiator!=acceptor && domain in initiator.domains && domain in acceptor.domains}
sig IProuteR extends RouteR{srcIP, dstIP: IPaddress}
	{domain in IPdomain && srcIP in initiator.id && dstIP in acceptor.id}
sig NDNrouteR extends RouteR{} 
	{domain in NDNdomain}
sig MFrouteR extends RouteR{srcGUID, dstGUID: NodeGUID}
	{domain in MFdomain && srcGUID in initiator.id && dstGUID in acceptor.id &&
	(some n:Node| n.id=dstGUID && n not in Repos => #dstGUID=1)}


one sig Connections{connected: Route -> Route, connectedR: RouteR -> RouteR}

fact NodeIDdomain{
	all n:Node, d:Domain |
		d in n.domains iff
			some nid:NodeID| nid in n.id &&
					(nid in IPaddress iff d in IPdomain) &&
					(nid in NDNclientID iff d in NDNdomain) &&
					(nid in NodeGUID iff d in MFdomain)
}

fact connected{
	all r1,r2:Route , c:Connections |
		(r1->r2) in c.connected iff
			r1.acceptor = r2.initiator && r1.contentID=r2.contentID && r1.reposdomain=r2.reposdomain
	all r1,r2:RouteR , c:Connections |
		(r1->r2) in c.connectedR iff
			r1.acceptor = r2.initiator && r1.content=r2.content && 
			(some gw:GW| r1.acceptor=gw && r2.initiator=gw =>
				(some n:NodeID, cid:ContentID| (n->cid) in gw.state1 && n in r2.acceptor.id &&
					(some re:Repos| some con:Content | (cid->con) in re.map 
					&& con in r1.content && con in r2.content))
				|| (some cid:ContentID| cid in gw.state2 &&
					(some re:Repos| some con:Content | (cid->con) in re.map 
					&& con in r1.content && con in r2.content))
					)
}

fact pathexists{
	all co:Connections, disj n1,n2:Node, cid:ContentID, rd: Domain| 
		(some r:Repos| rd in r.domains =>
		(some r1,r2:Route | 
		(r1->r2) in ^(co.connected) && r1.initiator=n1 && r2.acceptor=n2 && 
		r1.contentID=cid && r2.contentID=cid && r1.reposdomain=rd && r2.reposdomain=rd))

	all co:Connections, disj n1,n2:Node, con:Content, cd:Domain, rd:Domain| 
		(some c:Client, re:Repos| cd in c.domains && rd in re.domains =>
		(some r1,r2:RouteR | 
		(r1->r2) in ^(co.connectedR) && r1.initiator=n1 && r2.acceptor=n2 && 
		r1.content=con && r2.content=con))
}

fact NDNReversePath{
	all rr:NDNrouteR| 
		some r:NDNroute, re:Repos|(r.contentID -> rr.content) in re.map =>
				rr.initiator=r.acceptor && rr.acceptor=r.initiator && rr.domain=r.domain
				&& rr.content=re.map[r.contentID]
}


fact{
	#Repos=1 #Client=1
	all r:Repos, c:Client| no d:Domain| d in r.domains && d in c.domains
	all d:Domain| some gw:GW| d in gw.domains
	#GW=2
	all nid:NodeID | no disj n1,n2:Node | nid in n1.id && nid in n2.id
	//all dx:Demux | no disj n1,n2:Node | dx in n1.demux && dx in n2.demux
	//all disj r1,r2:Route| r1.initiator=r2.initiator => r1.demux!=r2.demux
	//all disj r1,r2:RouteR| r1.initiator=r2.initiator => r1.demux!=r2.demux
	#Domain=3
	#Client.want>=1
	//#Route>4
	//#RouteR>4
	#NDNroute>=1 #MFroute>=1 #IProute>=1
	#Route<8
	#RouteR<8
	//#NDNroute>=1 && #MFroute>=1 && #IProute>=1
	//#NDNrouteR>=1 && #MFrouteR>=1 && #IProuteR>=1
	//#GW.state>=1
}


//pred empty{}
//run empty for 1 Client, 2 GW, 1 Repos, 6 NodeID, 3 Domain, 1 ContentID, 1 Content, 12 Route, 12 RouteR, 1 Connections

pred reach[c:Client, cid:ContentID, re:Repos, gw:GW]{
	all co: Connections| cid in c.want =>
		(some r:Route, con:Content| r.initiator=c && r.acceptor=gw && r.contentID=cid
		=> some r1,r2:Route | (r1->r2) in ^(co.connected) && r1.initiator=gw && r2.acceptor=re
		&& r1.contentID=cid && r2.contentID=cid && (cid->con) in re.map
		&& r1.reposdomain in re.domains && r2.reposdomain in re.domains
		)
}
run reach for 1 Client, 2 GW, 1 Repos, 6 NodeID, 3 Domain, 1 ContentID, 1 Content, 12 Route, 12 RouteR, 1 Connections


assert reach{
	all c:Client, cid:ContentID| some re:Repos, gw:GW| reach[c,cid,re,gw]
}
check reach for 1 Client, 2 GW, 1 Repos, 6 NodeID, 3 Domain, 1 ContentID, 1 Content, 12 Route, 12 RouteR, 1 Connections

pred return[c:Client, cid:ContentID, re:Repos, gw:GW]{
	all co:Connections| some gw1:GW | reach[c,cid,re,gw1] =>
		(some r,r1,r2:RouteR| (r1->r2) in ^(co.connectedR) && r1.initiator=re && r2.acceptor=gw
			&& r1.content=re.map[cid] && r2.content=re.map[cid] && r.initiator=gw && r.acceptor=c 
			&& r.content=re.map[cid]
		)

}

run return for 1 Client, 2 GW, 1 Repos, 6 NodeID, 3 Domain, 1 ContentID, 1 Content, 12 Route, 12 RouteR, 1 Connections

assert return{
	all c:Client, cid:ContentID, re:Repos| some gw:GW| return[c,cid,re,gw]

}
check return  for 1 Client, 2 GW, 1 Repos, 6 NodeID, 3 Domain, 1 ContentID, 1 Content, 12 Route, 12 RouteR, 1 Connections
